"""Authors: Cody Baker and Ben Dichter."""
import spikeextractors as se

from ..basesortingextractorinterface import BaseSortingExtractorInterface


class TutorialSortingExtractor(se.NumpySortingExtractor):
    """Tutorial extractor for the toy data."""

    def __init__(self, duration: float = 10.0, K: int = 4, sampling_frequency: float = 30000.0):
        """
        Create tutorial recording extractor.

        Parameters
        ----------
        duration: float, optional
            Duration in seconds. Default 10 s.
        K: int, optional
            Number of spiking units. Default is 10.
        sampling_frequency: float, optional
            Sampling frequency. Default is 30000 Hz.
        """
        sorting = se.example_datasets.toy_example(duration=duration, K=K, sampling_frequency=sampling_frequency)[1]
        super().__init__()
        self.load_from_extractor(sorting=sorting)


class SortingTutorialInterface(BaseSortingExtractorInterface):
    """Sorting data interface for demonstrating NWB Conversion Tools usage in tutorials."""

    SX = TutorialSortingExtractor

    def __init__(
        self,
        duration: float = 10.0,
        num_units: int = 10,
        sampling_frequency: float = 30000.0,
        verbose: bool = True,
    ):
        """
        Initialize the internal properties of the recording interface.

        Parameters
        ----------
        duration: float, optional
            Duration in seconds. Default 10 s.
        num_units: int, optional
            Number of spiking units. Default is 10.
        sampling_frequency: float, optional
            Sampling frequency. Default is 30000 Hz.
        """
        super().__init__(
            duration=duration,
            K=num_units,  # You can handle spepcific argument renaming at this step of initializaiton
            sampling_frequency=sampling_frequency,
        )

        # Set data for custom columns of the Units table
        for unit_id in self.sorting_extractor.get_unit_ids():
            self.sorting_extractor.set_unit_property(
                unit_id=unit_id, property_name="custom_unit_column", value="A custom value"
            )
        self.verbose = verbose

    def get_metadata(self):
        # Set all automatically constructed metadata for the interface at this step
        # The user can always manually override this prior to running the conversion
        metadata = dict(
            Ecephys=dict(
                UnitProperties=[
                    dict(
                        name="custom_unit_column",
                        description="Custom column in the spiking unit table for the NWB Conversion Tools tutorial.",
                    )
                ]
            )
        )
        return metadata



"""Authors: Cody Baker and Ben Dichter."""
from pathlib import Path
from typing import Optional

import spikeextractors as se
from spikeinterface.core.old_api_utils import OldToNewRecording
from pynwb.ecephys import ElectricalSeries

from spikeinterface import BaseRecording
from spikeinterface.extractors import NeuroScopeRecordingExtractor, NeuroScopeSortingExtractor

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ..baselfpextractorinterface import BaseLFPExtractorInterface
from ..basesortingextractorinterface import BaseSortingExtractorInterface
from ....utils import FilePathType, FolderPathType, OptionalFilePathType, get_schema_from_hdmf_class, dict_deep_update
from .neuroscope_utils import get_session_start_time

try:
    import lxml
    from .neuroscope_utils import get_xml_file_path, get_channel_groups, get_shank_channels

    HAVE_LXML = True

except ImportError:
    HAVE_LXML = False
INSTALL_MESSAGE = "Please install lxml to use this interface!"


def subset_shank_channels(recording_extractor: BaseRecording, xml_file_path: str) -> BaseRecording:
    """Attempt to create a SubRecordingExtractor containing only channels related to neural data."""
    shank_channels = get_shank_channels(xml_file_path=xml_file_path)

    if shank_channels is not None:
        channel_ids = [channel_id for group in shank_channels for channel_id in group]
        new_ids = recording_extractor.get_channel_ids()[channel_ids]
        sub_recording = recording_extractor.channel_slice(new_ids)
    else:
        sub_recording = recording_extractor

    return sub_recording


def add_recording_extractor_properties(
    recording_extractor: BaseRecording, xml_file_path: str, gain: Optional[float] = None
) -> BaseRecording:
    """Automatically add properties to RecordingExtractor object."""

    if gain:
        recording_extractor.set_channel_gains(gain)

    channel_groups = get_channel_groups(xml_file_path=xml_file_path)

    channel_map = {
        channel_id: idx
        for idx, channel_id in enumerate([channel_id for group in channel_groups for channel_id in group])
    }
    group_electrode_numbers = [x for channels in channel_groups for x, _ in enumerate(channels)]
    group_nums = [n + 1 for n, channels in enumerate(channel_groups) for _ in channels]
    group_names = [f"Group{n}" for n in group_nums]

    channel_groups_mapped = [group_nums[channel_map[channel_id]] for channel_id in channel_map.keys()]
    group_names_mapped = [group_names[channel_map[channel_id]] for channel_id in channel_map.keys()]
    shank_electrode_number = [group_electrode_numbers[channel_map[channel_id]] for channel_id in channel_map.keys()]

    channel_ids_mapped = recording_extractor.get_channel_ids()
    recording_extractor.set_property(key="group", ids=channel_ids_mapped, values=channel_groups_mapped)
    recording_extractor.set_property(key="group_name", ids=channel_ids_mapped, values=group_names_mapped)
    recording_extractor.set_property(
        key="shank_electrode_number", ids=channel_ids_mapped, values=shank_electrode_number
    )


class NeuroscopeRecordingInterface(BaseRecordingExtractorInterface):
    """Primary data interface class for converting a NeuroscopeRecordingExtractor."""

    RX = NeuroScopeRecordingExtractor

    @staticmethod
    def get_ecephys_metadata(xml_file_path: str):
        """Auto-populates ecephys metadata from the xml_file_path."""
        channel_groups = get_channel_groups(xml_file_path=xml_file_path)
        ecephys_metadata = dict(
            ElectrodeGroup=[
                dict(
                    name=f"Group{n + 1}", description=f"Group{n + 1} electrodes.", location="", device="Device_ecephys"
                )
                for n, _ in enumerate(channel_groups)
            ],
            Electrodes=[
                dict(name="shank_electrode_number", description="0-indexed channel within a shank."),
                dict(name="group_name", description="The name of the ElectrodeGroup this electrode is a part of."),
            ],
        )
        return ecephys_metadata

    def __init__(
        self,
        file_path: FilePathType,
        gain: Optional[float] = None,
        xml_file_path: OptionalFilePathType = None,
        spikeextractors_backend: bool = False,
        verbose: bool = True,
    ):
        """
        Load and prepare raw acquisition data and corresponding metadata from the Neuroscope format (.dat files).

        Parameters
        ----------
        file_path : FilePathType
            Path to .dat file.
        gain : Optional[float], optional
            Conversion factors from int16 to Volts are not contained in xml_file_path; set them explicitly here.
            Most common value is 0.195 for an intan recording system.
            The default is None.
        xml_file_path : OptionalFilePathType, optional
            Path to .xml file containing device and electrode configuration.
            If unspecified, it will be automatically set as the only .xml file in the same folder as the .dat file.
            The default is None.
        spikeextractors_backend : bool
            False by default. When True the interface uses the old extractor from the spikextractors library instead
            of a new spikeinterface object.
        """
        assert HAVE_LXML, INSTALL_MESSAGE

        if xml_file_path is None:
            xml_file_path = get_xml_file_path(data_file_path=file_path)

        if spikeextractors_backend:
            self.RX = se.NeuroscopeRecordingExtractor
            super().__init__(file_path=file_path, xml_file_path=xml_file_path, verbose=verbose)
            self.recording_extractor = OldToNewRecording(oldapi_recording_extractor=self.recording_extractor)
        else:
            super().__init__(file_path=file_path, verbose=verbose)

        self.recording_extractor = subset_shank_channels(
            recording_extractor=self.recording_extractor, xml_file_path=xml_file_path
        )
        add_recording_extractor_properties(
            recording_extractor=self.recording_extractor, xml_file_path=xml_file_path, gain=gain
        )

    def get_metadata_schema(self):
        metadata_schema = super().get_metadata_schema()
        metadata_schema["properties"]["Ecephys"]["properties"].update(
            ElectricalSeries_raw=get_schema_from_hdmf_class(ElectricalSeries)
        )
        return metadata_schema

    def get_metadata(self):
        session_path = Path(self.source_data["file_path"]).parent
        session_id = session_path.stem
        xml_file_path = self.source_data.get("xml_file_path", str(session_path / f"{session_id}.xml"))
        metadata = super().get_metadata()
        metadata["Ecephys"].update(NeuroscopeRecordingInterface.get_ecephys_metadata(xml_file_path=xml_file_path))
        metadata["Ecephys"].update(
            ElectricalSeries_raw=dict(name="ElectricalSeries_raw", description="Raw acquisition traces.")
        )
        session_start_time = get_session_start_time(str(xml_file_path))
        if session_start_time is not None:
            metadata = dict_deep_update(metadata, dict(NWBFile=dict(session_start_time=session_start_time)))
        return metadata


class NeuroscopeMultiRecordingTimeInterface(NeuroscopeRecordingInterface):
    """Primary data interface class for converting a NeuroscopeMultiRecordingTimeExtractor."""

    RX = se.NeuroscopeMultiRecordingTimeExtractor

    def __init__(
        self,
        folder_path: FolderPathType,
        gain: Optional[float] = None,
        xml_file_path: OptionalFilePathType = None,
    ):
        """
        Load and prepare raw acquisition data and corresponding metadata from the Neuroscope format (.dat files).

        For all the .dat files in the folder_path, this concatenates them in time assuming no gaps in between.
        If there are gaps, timestamps inside the RecordingExtractor should be overridden.

        Parameters
        ----------
        folder_path : FolderPathType
            Path to folder of multiple .dat files.
        gain : Optional[float], optional
            Conversion factors from int16 to Volts are not contained in xml_file_path; set them explicitly here.
            Most common value is 0.195 for an intan recording system.
            The default is None.
        xml_file_path : OptionalFilePathType, optional
            Path to .xml file containing device and electrode configuration.
            If unspecified, it will be automatically set as the only .xml file in the same folder as the .dat file.
            The default is None.
        """
        assert HAVE_LXML, INSTALL_MESSAGE

        if xml_file_path is None:
            xml_file_path = get_xml_file_path(data_file_path=folder_path)
        super(NeuroscopeRecordingInterface, self).__init__(
            folder_path=folder_path,
            gain=gain,
            xml_file_path=xml_file_path,
        )
        self.recording_extractor = OldToNewRecording(oldapi_recording_extractor=self.recording_extractor)

        self.recording_extractor = subset_shank_channels(
            recording_extractor=self.recording_extractor, xml_file_path=xml_file_path
        )
        add_recording_extractor_properties(recording_extractor=self.recording_extractor, xml_file_path=xml_file_path)


class NeuroscopeLFPInterface(BaseLFPExtractorInterface):
    """Primary data interface class for converting Neuroscope LFP data."""

    RX = NeuroScopeRecordingExtractor

    def __init__(
        self,
        file_path: FilePathType,
        gain: Optional[float] = None,
        xml_file_path: OptionalFilePathType = None,
        spikeextractors_backend: bool = False,
    ):
        """
        Load and prepare lfp data and corresponding metadata from the Neuroscope format (.eeg or .lfp files).

        Parameters
        ----------
        file_path : FilePathType
            Path to .dat file.
        gain : Optional[float], optiona
            Conversion factors from int16 to Volts are not contained in xml_file_path; set them explicitly here.
            Most common value is 0.195 for an intan recording system.
            The default is None.
        xml_file_path : OptionalFilePathType, optional
            Path to .xml file containing device and electrode configuration.
            If unspecified, it will be automatically set as the only .xml file in the same folder as the .dat file.
            The default is None.
        spikeextractors_backend : bool
            False by default. When True the interface uses the old extractor from the spikextractors library instead
            of a new spikeinterface object.
        """
        assert HAVE_LXML, INSTALL_MESSAGE

        if xml_file_path is None:
            xml_file_path = get_xml_file_path(data_file_path=file_path)

        if spikeextractors_backend:
            self.RX = se.NeuroscopeRecordingExtractor
            super().__init__(file_path=file_path, xml_file_path=xml_file_path)
            self.recording_extractor = OldToNewRecording(oldapi_recording_extractor=self.recording_extractor)
        else:
            super().__init__(file_path=file_path)

        add_recording_extractor_properties(
            recording_extractor=self.recording_extractor, xml_file_path=xml_file_path, gain=gain
        )
        self.recording_extractor = subset_shank_channels(
            recording_extractor=self.recording_extractor, xml_file_path=xml_file_path
        )

    def get_metadata(self):
        session_path = Path(self.source_data["file_path"]).parent
        session_id = session_path.stem
        xml_file_path = self.source_data.get("xml_file_path", str(session_path / f"{session_id}.xml"))
        metadata = super().get_metadata()
        metadata["Ecephys"].update(NeuroscopeRecordingInterface.get_ecephys_metadata(xml_file_path=xml_file_path))
        return metadata


class NeuroscopeSortingInterface(BaseSortingExtractorInterface):
    """Primary data interface class for converting a NeuroscopeSortingExtractor."""

    SX = NeuroScopeSortingExtractor

    def __init__(
        self,
        folder_path: FolderPathType,
        keep_mua_units: bool = True,
        exclude_shanks: Optional[list] = None,
        xml_file_path: OptionalFilePathType = None,
        verbose: bool = True,
        spikeextractors_backend: bool = False,
        # TODO: we can enable this once
        #     a) waveforms on unit columns support conversion factor in NWB
        #     b) write_sorting utils support writing said waveforms properly to a units table
        # load_waveforms: bool = False,
        # gain: Optional[float] = None,
    ):
        """
        Load and prepare spike sorted data and corresponding metadata from the Neuroscope format (.res/.clu files).

        Parameters
        ----------
        folder_path : FolderPathType
            Path to folder containing .clu and .res files.
        keep_mua_units : bool
            Optional. Whether or not to return sorted spikes from multi-unit activity.
            The default is True.
        exclude_shanks : list
            Optional. List of indices to ignore. The set of all possible indices is chosen by default, extracted as the
            final integer of all the .res.%i and .clu.%i pairs.
        xml_file_path : OptionalFilePathType, optional
            Path to .xml file containing device and electrode configuration.
            If unspecified, it will be automatically set as the only .xml file in the same folder as the .dat file.
            The default is None.
        load_waveforms : bool, optional
            If True, extracts waveform data from .spk.%i files in the path corresponding to
            the .res.%i and .clue.%i files and sets these as unit spike features.
            The default is False.
            Not currently in use pending updates to NWB waveforms.
        gain : float, optional
            If loading waveforms, this value converts the data type of the waveforms to units of microvolts.
            Conversion factors from int16 to Volts are not contained in xml_file_path; set them explicitly here.
            Most common value is 0.195 for an intan recording system.
            The default is None.
            Not currently in use pending updates to NWB waveforms.
        spikeextractors_backend : bool
            False by default. When True the interface uses the old extractor from the spikextractors library instead
            of a new spikeinterface object.
        """
        assert HAVE_LXML, INSTALL_MESSAGE
        if spikeextractors_backend:
            self.SX = se.NeuroscopeMultiSortingExtractor

        super().__init__(
            folder_path=folder_path,
            keep_mua_units=keep_mua_units,
            exclude_shanks=exclude_shanks,
            xml_file_path=xml_file_path,
            verbose=verbose,
            # TODO: we can enable this once
            #     a) waveforms on unit columns support conversion factor in NWB
            #     b) write_sorting utils support writing said waveforms properly to a units table
            # load_waveforms=load_waveforms,
            # gain=gain,
        )

    def get_metadata(self):
        session_path = Path(self.source_data["folder_path"])
        session_id = session_path.stem
        xml_file_path = self.source_data.get("xml_file_path", str(session_path / f"{session_id}.xml"))
        metadata = dict(Ecephys=NeuroscopeRecordingInterface.get_ecephys_metadata(xml_file_path=xml_file_path))

        session_start_time = get_session_start_time(str(xml_file_path))
        if session_start_time is not None:
            metadata = dict_deep_update(metadata, dict(NWBFile=dict(session_start_time=session_start_time)))

        return metadata



"""Authors: Cody Baker and Ben Dichter."""
import spikeextractors as se

from pynwb.ecephys import ElectricalSeries

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ....utils import get_schema_from_hdmf_class


class TutorialRecordingExtractor(se.NumpyRecordingExtractor):
    """Tutorial extractor for the toy data."""

    def __init__(self, duration: float = 10.0, num_channels: int = 4, sampling_frequency: float = 30000.0):
        """
        Create tutorial recording extractor.

        Parameters
        ----------
        duration: float, optional
            Duration in seconds. Default 10 s.
        num_channels: int, optional
            Number of recording channels. Default is 4.
        sampling_frequency: float, optional
            Sampling frequency. Default is 30000 Hz.
        """
        recording = se.example_datasets.toy_example(
            duration=duration, num_channels=num_channels, sampling_frequency=sampling_frequency
        )[0]
        super().__init__(timeseries=recording._timeseries, sampling_frequency=sampling_frequency)


class RecordingTutorialInterface(BaseRecordingExtractorInterface):
    """High-pass recording data interface for demonstrating NWB Conversion Tools usage in tutorials."""

    RX = TutorialRecordingExtractor

    def __init__(self, duration: float = 10.0, num_channels: int = 4, sampling_frequency: float = 30000.0):
        """
        Initialize the internal properties of the recording interface.

        Parameters
        ----------
        duration: float, optional
            Duration in seconds. Default 10 s.
        num_channels: int, optional
            Number of recording channels. Default is 4.
        sampling_frequency: float, optional
            Sampling frequency. Default is 30000 Hz.
        """
        super().__init__(duration=duration, num_channels=num_channels, sampling_frequency=sampling_frequency)

        # Set manual group names at the recording extractor level
        for channel_id in self.recording_extractor.get_channel_ids():
            self.recording_extractor.set_channel_property(
                channel_id=channel_id, property_name="group_name", value="ElectrodeGroup"
            )
        # Set any other type of value that is intended to be written as a column to the final electrodes table
        for channel_id in self.recording_extractor.get_channel_ids():
            self.recording_extractor.set_channel_property(
                channel_id=channel_id, property_name="custom_electrodes_column", value="A custom value"
            )

    def get_metadata_schema(self):
        metadata_schema = super().get_metadata_schema()
        metadata_schema["properties"]["Ecephys"]["properties"].update(
            ElectricalSeries_raw=get_schema_from_hdmf_class(ElectricalSeries)
        )
        return metadata_schema

    def get_metadata(self):
        # Set all automatically constructed metadata for the interface at this step
        # The user can always manually override this prior to running the conversion
        metadata = dict(
            Ecephys=dict(
                Device=[dict(name="TutorialDevice", description="Device for the NWB Conversion Tools tutorial.")],
                ElectrodeGroup=[
                    dict(
                        name="ElectrodeGroup",
                        description="Electrode group for the NWB Conversion Tools tutorial.",
                        location="Somewhere in the brain",  # "unknown" if you don't know for sure
                        device="TutorialDevice",
                    )
                ],
                # Set the description metadata for the extra columns added during initialization
                Electrodes=[
                    dict(
                        name="group_name",
                        description="Custom ElectrodeGroup name for these electrodes.",
                    ),
                    dict(
                        name="custom_electrodes_column",
                        description="Custom column in the electrodes table for the NWB Conversion Tools tutorial.",
                    ),
                ],
                ElectricalSeries_raw=dict(
                    name="ElectricalSeries_raw",
                    description="Raw acquisition traces for the NWB Conversion Tools tutorial.",
                ),
            )
        )
        return metadata

    def get_conversion_options(self):
        conversion_options = dict(write_as="raw", es_key="ElectricalSeries_raw")
        return conversion_options



"""Authors: Cody Baker and Ben Dichter."""
from abc import abstractmethod, ABC
from typing import Optional

from pynwb import NWBFile

from .utils import get_base_schema, get_schema_from_method_signature


class BaseDataInterface(ABC):
    """Abstract class defining the structure of all DataInterfaces."""

    @classmethod
    def get_source_schema(cls):
        """Infer the JSON schema for the source_data from the method signature (annotation typing)."""
        return get_schema_from_method_signature(cls.__init__, exclude=["source_data"])

    @classmethod
    def get_conversion_options_schema(cls):
        """Infer the JSON schema for the conversion options from the method signature (annotation typing)."""
        return get_schema_from_method_signature(cls.run_conversion, exclude=["nwbfile", "metadata"])

    def __init__(self, **source_data):
        self.source_data = source_data

    def get_metadata_schema(self):
        """Retrieve JSON schema for metadata."""
        metadata_schema = get_base_schema(
            id_="metadata.schema.json",
            root=True,
            title="Metadata",
            description="Schema for the metadata",
            version="0.1.0",
        )
        return metadata_schema

    def get_metadata(self):
        """Child DataInterface classes should override this to match their metadata."""
        return dict()

    def get_conversion_options(self):
        """Child DataInterface classes should override this to match their conversion options."""
        return dict()

    @abstractmethod
    def run_conversion(
        self,
        nwbfile_path: Optional[str] = None,
        nwbfile: Optional[NWBFile] = None,
        metadata: Optional[dict] = None,
        overwrite: bool = False,
        **conversion_options,
    ):
        """
        Run the NWB conversion for the instantiated data interface.

        Parameters
        ----------
        nwbfile_path: FilePathType
            Path for where to write or load (if overwrite=False) the NWBFile.
            If specified, the context will always write to this location.
        nwbfile: NWBFile, optional
            An in-memory NWBFile object to write to the location.
        metadata: dict, optional
            Metadata dictionary with information used to create the NWBFile when one does not exist or overwrite=True.
        overwrite: bool, optional
            Whether or not to overwrite the NWBFile if one exists at the nwbfile_path.
            The default is False (append mode).
        verbose: bool, optional
            If 'nwbfile_path' is specified, informs user after a successful write operation.
            The default is True.
        """
        raise NotImplementedError("The run_conversion method for this DataInterface has not been defined!")



"""Authors: Heberto Mayorquin, Luiz Tauffer."""
from pathlib import Path
from platform import python_version
from sys import platform
from packaging import version

from spikeinterface.extractors import CedRecordingExtractor

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ....utils import get_schema_from_method_signature, FilePathType

HAVE_SONPY = True
try:
    import sonpy
except ImportError:
    HAVE_SONPY = False
INSTALL_MESSAGE = "Please install sonpy to use this interface (pip install sonpy)!"
if platform == "darwin" and version.parse(python_version()) < version.parse("3.8"):
    HAVE_SONPY = False
    INSTALL_MESSAGE = "The sonpy package (CED dependency) is not available on Mac for Python versions below 3.8!"


class CEDRecordingInterface(BaseRecordingExtractorInterface):
    """Primary data interface class for converting data from CED (Cambridge Electronic Design)
    ."""

    RX = CedRecordingExtractor

    @classmethod
    def get_source_schema(cls):
        source_schema = get_schema_from_method_signature(class_method=cls.__init__, exclude=["smrx_channel_ids"])
        source_schema.update(additionalProperties=True)
        source_schema["properties"]["file_path"].update(description="Path to CED data file.")
        return source_schema

    @classmethod
    def get_all_channels_info(cls, file_path: FilePathType):
        """Retrieve and inspect necessary channel information prior to initialization."""
        assert HAVE_SONPY, INSTALL_MESSAGE
        return cls.RX.get_all_channels_info(file_path=file_path)

    def __init__(self, file_path: FilePathType, verbose: bool = True):
        assert HAVE_SONPY, INSTALL_MESSAGE
        stream_id = None
        if Path(file_path).suffix == ".smr":
            stream_id = "1"
        super().__init__(file_path=file_path, stream_id=stream_id, verbose=verbose)



import json
from dateutil.parser import parse as dateparse
from typing import Optional

from roiextractors import ScanImageTiffImagingExtractor

try:
    from ScanImageTiffReader import ScanImageTiffReader

    HAVE_SCAN_IMAGE_TIFF = True
except ImportError:
    HAVE_SCAN_IMAGE_TIFF = False


from ..baseimagingextractorinterface import BaseImagingExtractorInterface
from ....utils import FilePathType


def extract_extra_metadata(file_path):

    description = ScanImageTiffReader(str(file_path)).description(iframe=0)
    extra_metadata = {x.split("=")[0]: x.split("=")[1] for x in description.split("\r") if "=" in x}

    return extra_metadata


class ScanImageImagingInterface(BaseImagingExtractorInterface):

    IX = ScanImageTiffImagingExtractor

    @classmethod
    def get_source_schema(cls):
        source_schema = super().get_source_schema()
        source_schema["properties"]["file_path"]["description"] = "Path to Tiff file."
        return source_schema

    def __init__(
        self,
        file_path: FilePathType,
        fallback_sampling_frequency: Optional[float] = None,
        verbose: bool = True,
    ):
        """
        DataInterface for reading Tiff files that are generated by ScanImage. This interface extracts the metadata
        from the exif of the tiff file.

        Parameters
        ----------
        file_path: str
            Path to tiff file.
        fallback_sampling_frequency: float, optional
            The sampling frequency can usually be extracted from the scanimage metadata in
            exif:ImageDescription:state.acq.frameRate. If not, use this.
        """

        assert (
            HAVE_SCAN_IMAGE_TIFF
        ), "To use the ScanImageTiffExtractor install scanimage-tiff-reader: \n\n pip install scanimage-tiff-reader\n\n"
        self.image_metadata = extract_extra_metadata(file_path=file_path)

        if "state.acq.frameRate" in self.image_metadata:
            sampling_frequency = float(self.image_metadata["state.acq.frameRate"])
        else:
            assert_msg = (
                "sampling frequency not found in image metadata, "
                "input the frequency using the argument `fallback_sampling_frequency`"
            )
            assert fallback_sampling_frequency is not None, assert_msg
            sampling_frequency = fallback_sampling_frequency

        super().__init__(file_path=file_path, sampling_frequency=sampling_frequency, verbose=verbose)

    def get_metadata(self):
        device_number = 0  # Imaging plane metadata is a list with metadata for each plane

        metadata = super().get_metadata()

        if "state.internal.triggerTimeString" in self.image_metadata:
            extracted_session_start_time = dateparse(self.image_metadata["state.internal.triggerTimeString"])
            metadata["NWBFile"] = dict(session_start_time=extracted_session_start_time)

        # Extract many scan image properties and attach them as dic in the description
        ophys_metadata = metadata["Ophys"]
        two_photon_series_metadata = ophys_metadata["TwoPhotonSeries"][device_number]
        if self.image_metadata is not None:
            extracted_description = json.dumps(self.image_metadata)
            two_photon_series_metadata.update(description=extracted_description)

        return metadata



from roiextractors import Hdf5ImagingExtractor

from ..baseimagingextractorinterface import BaseImagingExtractorInterface
from ....utils import FilePathType, FloatType, ArrayType


class Hdf5ImagingInterface(BaseImagingExtractorInterface):
    """Data Interface for Hdf5ImagingExtractor."""

    IX = Hdf5ImagingExtractor

    def __init__(
        self,
        file_path: FilePathType,
        mov_field: str = "mov",
        sampling_frequency: FloatType = None,
        start_time: FloatType = None,
        metadata: dict = None,
        channel_names: ArrayType = None,
        verbose: bool = True,
    ):
        super().__init__(
            file_path=file_path,
            mov_field=mov_field,
            sampling_frequency=sampling_frequency,
            start_time=start_time,
            metadata=metadata,
            channel_names=channel_names,
            verbose=verbose,
        )



"""Author: Luiz Tauffer."""
from abc import ABC
from typing import Optional

from pynwb import NWBFile, NWBHDF5IO
from pynwb.device import Device
from pynwb.icephys import IntracellularElectrode

from ...basedatainterface import BaseDataInterface
from ...tools.neo import (
    get_number_of_electrodes,
    get_number_of_segments,
    write_neo_to_nwb,
)
from ...tools.nwb_helpers import make_nwbfile_from_metadata
from ...utils import (
    OptionalFilePathType,
    get_schema_from_hdmf_class,
    get_schema_from_method_signature,
    get_base_schema,
    get_metadata_schema_for_icephys,
)


try:
    from ndx_dandi_icephys import DandiIcephysMetadata

    HAVE_NDX_DANDI_ICEPHYS = True
except ImportError:
    DandiIcephysMetadata = None
    HAVE_NDX_DANDI_ICEPHYS = False


class BaseIcephysInterface(BaseDataInterface, ABC):
    """Primary class for all intracellular NeoInterfaces."""

    neo_class = None

    @classmethod
    def get_source_schema(cls):
        source_schema = get_schema_from_method_signature(class_method=cls.__init__, exclude=[])
        return source_schema

    def __init__(self, file_paths: list):
        super().__init__(file_paths=file_paths)

        self.readers_list = list()
        for f in file_paths:
            self.readers_list.append(self.neo_class(filename=f))

        self.subset_channels = None
        self.n_segments = get_number_of_segments(neo_reader=self.readers_list[0], block=0)
        self.n_channels = get_number_of_electrodes(neo_reader=self.readers_list[0])

    def get_metadata_schema(self):
        metadata_schema = super().get_metadata_schema()
        if DandiIcephysMetadata:
            metadata_schema["properties"]["ndx-dandi-icephys"] = get_schema_from_hdmf_class(DandiIcephysMetadata)
        metadata_schema["properties"]["Icephys"] = get_metadata_schema_for_icephys()
        return metadata_schema

    def get_metadata(self):
        metadata = super().get_metadata()
        metadata["Icephys"] = dict(
            Device=[dict(name="DeviceIcephys", description="no description")],
            Electrode=[
                dict(name=f"electrode-{i}", description="no description", device="DeviceIcephys")
                for i in range(get_number_of_electrodes(self.readers_list[0]))
            ],
        )
        return metadata

    def run_conversion(
        self,
        nwbfile: NWBFile = None,
        metadata: dict = None,
        stub_test: bool = False,
        save_path: OptionalFilePathType = None,
        overwrite: bool = False,
        icephys_experiment_type: Optional[str] = None,
        skip_electrodes: tuple = (),
    ):
        """
        Primary function for converting raw (unprocessed) RecordingExtractor data to the NWB standard.

        Parameters
        ----------
        nwbfile: NWBFile
            nwb file to which the recording information is to be added
        metadata: dict
            metadata info for constructing the nwb file (optional).
            Should be of the format
                metadata['Ecephys']['ElectricalSeries'] = dict(name=my_name, description=my_description)
        use_times: bool
            If True, the times are saved to the nwb file using recording.frame_to_time(). If False (default),
            the sampling rate is used.
        save_path: PathType
            Required if an nwbfile is not passed. Must be the path to the nwbfile
            being appended, otherwise one is created and written.
        overwrite: bool
            If using save_path, whether or not to overwrite the NWBFile if it already exists.
        stub_test: bool, optional (default False)
            If True, will truncate the data to run the conversion faster and take up less memory.
        icephys_experiment_type: str (optional)
            Type of Icephys experiment. Allowed types are: 'voltage_clamp', 'current_clamp' and 'izero' (all current and amplifier settings turned off).
            If no value is passed, 'voltage_clamp' is used as default.
        skip_electrodes: tuple

        """
        if nwbfile is None:
            nwbfile = make_nwbfile_from_metadata(metadata)

        if (
            HAVE_NDX_DANDI_ICEPHYS
            and "ndx-dandi-icephys" in metadata
            and "DandiIcephysMetadata" not in nwbfile.lab_meta_data
        ):
            nwbfile.add_lab_meta_data(DandiIcephysMetadata(**metadata["ndx-dandi-icephys"]))

        for i, reader in enumerate(self.readers_list):
            write_neo_to_nwb(
                neo_reader=reader,
                nwbfile=nwbfile,
                metadata=metadata,
                overwrite=overwrite if i == 0 else False,
                icephys_experiment_type=metadata["Icephys"]["Sessions"][i]["icephys_experiment_type"],
                stimulus_type=metadata["Icephys"]["Sessions"][i]["stimulus_type"],
                skip_electrodes=skip_electrodes,
            )

        if save_path:
            with NWBHDF5IO(save_path, "w") as io:
                io.write(nwbfile)



"""Authors: Cody Baker and Ben Dichter."""
from abc import ABC
from typing import Optional

import numpy as np
import spikeextractors as se
import spikeinterface as si
from pynwb import NWBFile
from pynwb.device import Device
from pynwb.ecephys import ElectrodeGroup

from ...basedatainterface import BaseDataInterface
from ...tools.spikeinterface import write_recording
from ...utils import get_schema_from_hdmf_class, get_base_schema, OptionalFilePathType


class BaseRecordingExtractorInterface(BaseDataInterface, ABC):
    """Primary class for all RecordingExtractorInterfaces."""

    RX = None

    def __init__(self, verbose: bool = True, **source_data):
        super().__init__(**source_data)
        self.recording_extractor = self.RX(**source_data)
        self.subset_channels = None
        self.verbose = verbose

    def get_metadata_schema(self):
        """Compile metadata schema for the RecordingExtractor."""
        metadata_schema = super().get_metadata_schema()
        metadata_schema["properties"]["Ecephys"] = get_base_schema(tag="Ecephys")
        metadata_schema["properties"]["Ecephys"]["required"] = ["Device", "ElectrodeGroup"]
        metadata_schema["properties"]["Ecephys"]["properties"] = dict(
            Device=dict(type="array", minItems=1, items={"$ref": "#/properties/Ecephys/properties/definitions/Device"}),
            ElectrodeGroup=dict(
                type="array", minItems=1, items={"$ref": "#/properties/Ecephys/properties/definitions/ElectrodeGroup"}
            ),
            Electrodes=dict(
                type="array",
                minItems=0,
                renderForm=False,
                items={"$ref": "#/properties/Ecephys/properties/definitions/Electrodes"},
            ),
        )
        # Schema definition for arrays
        metadata_schema["properties"]["Ecephys"]["properties"]["definitions"] = dict(
            Device=get_schema_from_hdmf_class(Device),
            ElectrodeGroup=get_schema_from_hdmf_class(ElectrodeGroup),
            Electrodes=dict(
                type="object",
                additionalProperties=False,
                required=["name"],
                properties=dict(
                    name=dict(type="string", description="name of this electrodes column"),
                    description=dict(type="string", description="description of this electrodes column"),
                ),
            ),
        )
        return metadata_schema

    def get_metadata(self):
        metadata = super().get_metadata()
        metadata["Ecephys"] = dict(
            Device=[dict(name="Device_ecephys", description="no description")],
            ElectrodeGroup=[
                dict(name=str(group_id), description="no description", location="unknown", device="Device_ecephys")
                for group_id in np.unique(self.recording_extractor.get_channel_groups())
            ],
        )
        return metadata

    def subset_recording(self, stub_test: bool = False):
        """
        Subset a recording extractor according to stub and channel subset options.

        Parameters
        ----------
        stub_test : bool, optional (default False)
        """
        kwargs = dict()

        if stub_test:
            num_frames = 100
            end_frame = min([num_frames, self.recording_extractor.get_num_frames()])
            kwargs.update(end_frame=end_frame)
        if self.subset_channels is not None:
            kwargs.update(channel_ids=self.subset_channels)
        if isinstance(self.recording_extractor, se.RecordingExtractor):
            recording_extractor = se.SubRecordingExtractor(self.recording_extractor, **kwargs)
        elif isinstance(self.recording_extractor, si.BaseRecording):
            recording_extractor = self.recording_extractor.frame_slice(start_frame=0, end_frame=end_frame)
        else:
            raise TypeError(f"{self.recording_extractor} should be either se.RecordingExtractor or si.BaseRecording")
        return recording_extractor

    def run_conversion(
        self,
        nwbfile_path: OptionalFilePathType = None,
        nwbfile: Optional[NWBFile] = None,
        metadata: Optional[dict] = None,
        overwrite: bool = False,
        stub_test: bool = False,
        starting_time: Optional[float] = None,
        use_times: bool = False,
        write_as: Optional[str] = None,
        write_electrical_series: bool = True,
        es_key: str = None,
        compression: Optional[str] = None,
        compression_opts: Optional[int] = None,
        iterator_type: Optional[str] = None,
        iterator_opts: Optional[dict] = None,
        save_path: OptionalFilePathType = None,  # TODO: to be removed, depreceation applied at tools level
    ):
        """
        Primary function for converting raw (unprocessed) RecordingExtractor data to the NWB standard.

        Parameters
        ----------
        nwbfile_path: FilePathType
            Path for where to write or load (if overwrite=False) the NWBFile.
            If specified, this context will always write to this location.
        nwbfile: NWBFile
            nwb file to which the recording information is to be added
        metadata: dict
            metadata info for constructing the nwb file (optional).
            Should be of the format::

                metadata['Ecephys']['ElectricalSeries'] = dict(name=my_name, description=my_description)
        overwrite: bool, optional
            Whether or not to overwrite the NWBFile if one exists at the nwbfile_path.
        The default is False (append mode).
        starting_time: float (optional)
            Sets the starting time of the ElectricalSeries to a manually set value.
            Increments timestamps if use_times is True.
        use_times: bool
            If True, the times are saved to the nwb file using recording.frame_to_time(). If False (default),
            the sampling rate is used.
        stub_test: bool, optional (default False)
            If True, will truncate the data to run the conversion faster and take up less memory.
        write_as: str (optional, defaults to 'raw')
            Options: 'raw', 'lfp' or 'processed'
        write_electrical_series: bool (optional)
            If True (default), electrical series are written in acquisition. If False, only device, electrode_groups,
            and electrodes are written to NWB.
        es_key: str (optional)
            Key in metadata dictionary containing metadata info for the specific electrical series
        compression: str (optional, defaults to "gzip")
            Type of compression to use. Valid types are "gzip" and "lzf".
            Set to None to disable all compression.
        compression_opts: int (optional, defaults to 4)
            Only applies to compression="gzip". Controls the level of the GZIP.
        iterator_type: str (optional, defaults to 'v2')
            The type of DataChunkIterator to use.
            'v1' is the original DataChunkIterator of the hdmf data_utils.
            'v2' is the locally developed RecordingExtractorDataChunkIterator, which offers full control over chunking.
        iterator_opts: dict (optional)
            Dictionary of options for the RecordingExtractorDataChunkIterator (iterator_type='v2').
            Valid options are
                buffer_gb : float (optional, defaults to 1 GB)
                    Recommended to be as much free RAM as available). Automatically calculates suitable buffer shape.
                chunk_mb : float (optional, defaults to 1 MB)
                    Should be below 1 MB. Automatically calculates suitable chunk shape.
            If manual specification of buffer_shape and chunk_shape are desired, these may be specified as well.
        """
        if stub_test or self.subset_channels is not None:
            recording = self.subset_recording(stub_test=stub_test)
        else:
            recording = self.recording_extractor

        write_recording(
            recording=recording,
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            metadata=metadata,
            overwrite=overwrite,
            verbose=self.verbose,
            starting_time=starting_time,
            use_times=use_times,
            write_as=write_as,
            write_electrical_series=write_electrical_series,
            es_key=es_key,
            compression=compression,
            compression_opts=compression_opts,
            iterator_type=iterator_type,
            iterator_opts=iterator_opts,
            save_path=save_path,  # TODO: to be removed, depreceation applied at tools level
        )



"""Authors: Cody Baker and Ben Dichter."""
from pathlib import Path

import spikeextractors as se
import numpy as np

from ..basesortingextractorinterface import BaseSortingExtractorInterface
from ....utils import FilePathType

try:
    import scipy.io
    import hdf5storage

    HAVE_SCIPY_AND_HDF5STORAGE = True
except ImportError:
    HAVE_SCIPY_AND_HDF5STORAGE = False


class CellExplorerSortingInterface(BaseSortingExtractorInterface):
    """Primary data interface class for converting Cell Explorer spiking data."""

    SX = se.CellExplorerSortingExtractor

    def __init__(self, file_path: FilePathType, verbose: bool = True):
        super().__init__(spikes_matfile_path=file_path, verbose=verbose)
        self.source_data = dict(file_path=file_path)
        spikes_matfile_path = Path(file_path)

        session_path = Path(file_path).parent
        session_id = session_path.stem

        assert (
            spikes_matfile_path.is_file()
        ), f"The file_path should point to an existing .spikes.cellinfo.mat file ({spikes_matfile_path})"

        try:
            spikes_mat = scipy.io.loadmat(file_name=str(spikes_matfile_path))
            self.read_spikes_info_with_scipy = True
        except NotImplementedError:
            spikes_mat = hdf5storage.loadmat(file_name=str(spikes_matfile_path))
            self.read_spikes_info_with_scipy = False
        cell_info = spikes_mat.get("spikes", np.empty(0))
        self.cell_info_fields = cell_info.dtype.names

        if self.read_spikes_info_with_scipy:
            unit_ids = self.sorting_extractor.get_unit_ids()
            if "cluID" in self.cell_info_fields:
                for unit_id, value in zip(unit_ids, [int(x) for x in cell_info["cluID"][0][0][0]]):
                    self.sorting_extractor.set_unit_property(unit_id=unit_id, property_name="clu_id", value=value)
            if "shankID" in self.cell_info_fields:
                for unit_id, value in zip(unit_ids, [f"Group{x}" for x in cell_info["shankID"][0][0][0]]):
                    self.sorting_extractor.set_unit_property(unit_id=unit_id, property_name="group_id", value=value)
            if "region" in self.cell_info_fields:
                for unit_id, value in zip(unit_ids, [str(x[0]) for x in cell_info["region"][0][0][0]]):
                    self.sorting_extractor.set_unit_property(unit_id=unit_id, property_name="location", value=value)
        else:  # Logic for hdf5storage
            unit_ids = self.sorting_extractor.get_unit_ids()
            if "cluID" in self.cell_info_fields:
                for unit_id, value in zip(unit_ids, [int(x) for x in cell_info["cluID"][0][0]]):
                    self.sorting_extractor.set_unit_property(unit_id=unit_id, property_name="clu_id", value=value)
            if "shankID" in self.cell_info_fields:
                for unit_id, value in zip(unit_ids, [f"Group{x}" for x in cell_info["shankID"][0][0]]):
                    self.sorting_extractor.set_unit_property(unit_id=unit_id, property_name="group_id", value=value)
            if "region" in self.cell_info_fields:
                for unit_id, value in zip(unit_ids, [str(x[0]) for x in cell_info["region"][0]][0]):
                    self.sorting_extractor.set_unit_property(unit_id=unit_id, property_name="location", value=value)
        celltype_mapping = {"pE": "excitatory", "pI": "inhibitory", "[]": "unclassified"}
        celltype_file_path = session_path / f"{session_id}.CellClass.cellinfo.mat"
        if celltype_file_path.is_file():
            celltype_info = scipy.io.loadmat(celltype_file_path).get("CellClass", np.empty(0))
            if "label" in celltype_info.dtype.names:
                for unit_id, value in zip(
                    unit_ids, [str(celltype_mapping[str(x[0])]) for x in celltype_info["label"][0][0][0]]
                ):
                    self.sorting_extractor.set_unit_property(unit_id=unit_id, property_name="cell_type", value=value)

    def get_metadata(self):
        session_path = Path(self.source_data["file_path"]).parent
        session_id = session_path.stem
        # TODO: add condition for retrieving ecephys metadata if no recording or lfp are included in conversion
        metadata = dict(NWBFile=dict(session_id=session_id))

        unit_properties = []
        cellinfo_file_path = session_path / f"{session_id}.spikes.cellinfo.mat"
        if cellinfo_file_path.is_file():
            cell_info_fields = self.cell_info_fields
            if "cluID" in cell_info_fields:
                unit_properties.append(
                    dict(
                        name="clu_id",
                        description="0-indexed id of cluster identified from the shank.",
                    )
                )
            if "shankID" in cell_info_fields:
                unit_properties.append(
                    dict(
                        name="group_id",
                        description="The electrode group ID that each unit was identified by.",
                    )
                )
            if "region" in cell_info_fields:
                unit_properties.append(
                    dict(
                        name="location",
                        description="Brain region where each unit was detected.",
                    )
                )
        celltype_filepath = session_path / f"{session_id}.CellClass.cellinfo.mat"
        if celltype_filepath.is_file():
            celltype_info = scipy.io.loadmat(celltype_filepath).get("CellClass", np.empty(0))
            if "label" in celltype_info.dtype.names:
                unit_properties.append(
                    dict(
                        name="cell_type",
                        description="Type of cell this has been classified as.",
                    )
                )
        metadata.update(Ecephys=dict(UnitProperties=unit_properties))
        return metadata



from .types import (
    FilePathType,
    FolderPathType,
    OptionalFilePathType,
    OptionalFolderPathType,
    ArrayType,
    OptionalArrayType,
    FloatType,
    IntType,
)
from .dict import dict_deep_update, load_dict_from_file, append_replace_dict_in_list, exist_dict_in_list
from .json_schema import (
    get_schema_from_method_signature,
    get_schema_from_hdmf_class,
    get_schema_for_NWBFile,
    get_base_schema,
    unroot_schema,
    fill_defaults,
    get_metadata_schema_for_icephys,
)
from .globbing import decompose_f_string, parse_f_string

from .checks import calculate_regular_series_rate



from datetime import datetime
from pathlib import Path

import spikeextractors as se
from ....utils import FilePathType


def _assert_single_shank_for_spike_extractors(recording: se.SpikeGLXRecordingExtractor):
    """Raises an exception for a se.SpikeGLXRecordingExtractor object intialized in a file
    with complex geometry as this is not (and will not be )supported in the old spikeextractors API.

    Parameters
    ----------
    recording : se.SpikeGLXRecordingExtractor
        a newly instantiated version of the spikeextractors object

    Raises
    ------
    NotImplementedError
        Raises a not implemented error.
    """
    meta = recording._meta
    # imDatPrb_type 0 and 21 correspond to single shank channels
    # see https://billkarsh.github.io/SpikeGLX/help/imroTables/
    imDatPrb_type = meta["imDatPrb_type"]
    if imDatPrb_type not in ["0", "21"]:
        raise NotImplementedError(
            "More than a single shank is not supported in spikeextractors, use the new spikeinterface."
        )


def _fetch_metadata_dic_for_spikextractors_spikelgx_object(recording: se.SpikeGLXRecordingExtractor) -> dict:
    """
    fetches the meta file from a se.SpikeGLXRecordingExtractor object.
    Parameters
    ----------
    recording : se.SpikeGLXRecordingExtractor
        a newly instantiated version of the spikeextractors object


    Returns
    -------
    dict
        a dictionary with the metadadata concerning the recording
    """

    if isinstance(recording, se.SubRecordingExtractor):
        recording_metadata = recording._parent_recording._meta
    else:
        recording_metadata = recording._meta

    return recording_metadata


def get_session_start_time(recording_metadata: dict) -> datetime:
    """Fetches the session start time from the recording_metadata dic
    Parameters
    ----------
    recording_metadata : dict
        the metadata dic as obtained from the Spikelgx recording.

    Returns
    -------
    datetime
        the session start time in datetime format.
    """

    session_start_time = recording_metadata.get("fileCreateTime", None)
    if session_start_time:
        session_start_time = datetime.fromisoformat(session_start_time)

    return session_start_time


def fetch_stream_id_for_spikelgx_file(file_path: FilePathType) -> str:
    """Returns the stream_id for a spikelgx file

    Example of file name structure:
    Consider the filenames: `Noise4Sam_g0_t0.nidq.bin` or `Noise4Sam_g0_t0.imec0.lf.bin`
    The filenames consist of 3 or 4 parts separated by `.`
    1. "Noise4Sam_g0_t0" will be the `name` variable. This choosen by the user at recording time.
    2. "_gt0_" will give the `seg_index` (here 0)
    3. "nidq" or "imec0" will give the `device` variable
    4. "lf" or "ap" will be the `signal_kind` variable (for nidq the signal kind is an empty string)

    stream_id is the concatenation of `device.signal_kind`

    Parameters
    ----------
    file_path : FilePathType
        The file_path of spikelgx file.

    Returns
    -------
    str
        the stream_id
    """

    suffixes = Path(file_path).suffixes
    device = next(suffix for suffix in suffixes if "imec" in suffix or "nidq" in suffix)
    signal_kind = ""
    if "imec" in device:
        signal_kind = next(suffix for suffix in suffixes if "ap" in suffix or "lf" in suffix)

    stream_id = device[1:] + signal_kind

    return stream_id



from roiextractors import CaimanSegmentationExtractor

from ..basesegmentationextractorinterface import BaseSegmentationExtractorInterface
from ....utils import FilePathType


class CaimanSegmentationInterface(BaseSegmentationExtractorInterface):
    """Data interface for CaimanSegmentationExtractor."""

    SegX = CaimanSegmentationExtractor

    def __init__(self, file_path: FilePathType):
        super().__init__(file_path=file_path)



from .ecephys.tutorial.recordingtutorialdatainterface import RecordingTutorialInterface
from .ecephys.tutorial.sortingtutorialdatainterface import SortingTutorialInterface
from .ecephys.neuroscope.neuroscopedatainterface import (
    NeuroscopeRecordingInterface,
    NeuroscopeLFPInterface,
    NeuroscopeMultiRecordingTimeInterface,
    NeuroscopeSortingInterface,
)
from .ecephys.spikeglx.spikeglxdatainterface import SpikeGLXRecordingInterface, SpikeGLXLFPInterface
from .ecephys.spikegadgets.spikegadgetsdatainterface import SpikeGadgetsRecordingInterface
from .ecephys.spikeinterface.sipickledatainterfaces import (
    SIPickleRecordingExtractorInterface,
    SIPickleSortingExtractorInterface,
)
from .ecephys.intan.intandatainterface import IntanRecordingInterface
from .ecephys.ced.ceddatainterface import CEDRecordingInterface
from .ecephys.cellexplorer.cellexplorerdatainterface import CellExplorerSortingInterface
from .ecephys.blackrock.blackrockdatainterface import (
    BlackrockRecordingExtractorInterface,
    BlackrockSortingExtractorInterface,
)
from .ecephys.openephys.openephysdatainterface import (
    OpenEphysRecordingExtractorInterface,
    OpenEphysSortingExtractorInterface,
)
from .ecephys.axona.axonadatainterface import (
    AxonaRecordingExtractorInterface,
    AxonaPositionDataInterface,
    AxonaLFPDataInterface,
    AxonaUnitRecordingExtractorInterface,
)
from .ecephys.neuralynx.neuralynxdatainterface import NeuralynxRecordingInterface
from .ecephys.phy.phydatainterface import PhySortingInterface
from .ecephys.kilosort.kilosortdatainterface import KilosortSortingInterface

from .ophys.caiman.caimandatainterface import CaimanSegmentationInterface
from .ophys.cnmfe.cnmfedatainterface import CnmfeSegmentationInterface
from .ophys.suite2p.suite2pdatainterface import Suite2pSegmentationInterface
from .ophys.extract.extractdatainterface import ExtractSegmentationInterface
from .ophys.sima.simadatainterface import SimaSegmentationInterface

from .ophys.sbx.sbxdatainterface import SbxImagingInterface
from .ophys.tiff.tiffdatainterface import TiffImagingInterface
from .ophys.hdf5.hdf5datainterface import Hdf5ImagingInterface
from .ophys.scanimage.scanimageimaginginterface import ScanImageImagingInterface

from .behavior.movie.moviedatainterface import MovieInterface

from .icephys.abf.abfdatainterface import AbfInterface


interface_list = [
    RecordingTutorialInterface,
    SortingTutorialInterface,
    NeuralynxRecordingInterface,
    NeuroscopeRecordingInterface,
    NeuroscopeMultiRecordingTimeInterface,
    NeuroscopeSortingInterface,
    NeuroscopeLFPInterface,
    SpikeGLXRecordingInterface,
    SpikeGLXLFPInterface,
    SpikeGadgetsRecordingInterface,
    SIPickleRecordingExtractorInterface,
    SIPickleSortingExtractorInterface,
    IntanRecordingInterface,
    CEDRecordingInterface,
    CellExplorerSortingInterface,
    BlackrockRecordingExtractorInterface,
    BlackrockSortingExtractorInterface,
    OpenEphysRecordingExtractorInterface,
    OpenEphysSortingExtractorInterface,
    PhySortingInterface,
    KilosortSortingInterface,
    AxonaRecordingExtractorInterface,
    AxonaPositionDataInterface,
    AxonaLFPDataInterface,
    AxonaUnitRecordingExtractorInterface,
    CaimanSegmentationInterface,
    CnmfeSegmentationInterface,
    Suite2pSegmentationInterface,
    ExtractSegmentationInterface,
    SimaSegmentationInterface,
    SbxImagingInterface,
    TiffImagingInterface,
    Hdf5ImagingInterface,
    MovieInterface,
    AbfInterface,
    ScanImageImagingInterface,
]



"""Authors: Alessio Buccino."""
from spikeextractors import load_extractor_from_pickle

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ..basesortingextractorinterface import BaseSortingExtractorInterface
from ....utils import FilePathType


class SIPickleRecordingExtractorInterface(BaseRecordingExtractorInterface):
    """Primary interface for reading and converting SpikeInterface Recording objects through .pkl files."""

    RX = None

    def __init__(self, file_path: FilePathType, verbose: bool = True):
        self.recording_extractor = load_extractor_from_pickle(pkl_file=file_path)
        self.subset_channels = None
        self.source_data = dict(file_path=file_path)
        self.verbose = verbose


class SIPickleSortingExtractorInterface(BaseSortingExtractorInterface):
    """Primary interface for reading and converting SpikeInterface Sorting objects through .pkl files."""

    SX = None

    def __init__(self, file_path: FilePathType, verbose: bool = True):
        self.sorting_extractor = load_extractor_from_pickle(pkl_file=file_path)
        self.source_data = dict(file_path=file_path)
        self.verbose = verbose



"""Authors: Cody Baker and Ben Dichter."""
from abc import ABC

from pynwb import NWBFile
from pynwb.device import Device
from pynwb.ophys import Fluorescence, ImageSegmentation, ImagingPlane, TwoPhotonSeries

from ...basedatainterface import BaseDataInterface
from ...tools.roiextractors import write_segmentation, get_nwb_segmentation_metadata
from ...utils import (
    get_schema_from_hdmf_class,
    fill_defaults,
    get_base_schema,
)


class BaseSegmentationExtractorInterface(BaseDataInterface, ABC):
    SegX = None

    def __init__(self, **source_data):
        super().__init__(**source_data)
        self.segmentation_extractor = self.SegX(**source_data)

    def get_metadata_schema(self):
        metadata_schema = super().get_metadata_schema()
        metadata_schema["required"] = ["Ophys"]
        metadata_schema["properties"]["Ophys"] = get_base_schema()
        metadata_schema["properties"]["Ophys"]["properties"] = dict(
            Device=dict(type="array", minItems=1, items=get_schema_from_hdmf_class(Device)),
        )
        metadata_schema["properties"]["Ophys"]["properties"].update(
            Fluorescence=get_schema_from_hdmf_class(Fluorescence),
            ImageSegmentation=get_schema_from_hdmf_class(ImageSegmentation),
            ImagingPlane=get_schema_from_hdmf_class(ImagingPlane),
            TwoPhotonSeries=get_schema_from_hdmf_class(TwoPhotonSeries),
        )
        metadata_schema["properties"]["Ophys"]["required"] = ["Device", "Fluorescence", "ImageSegmentation"]

        # Temporary fixes until centralized definition of metadata schemas
        metadata_schema["properties"]["Ophys"]["properties"]["ImagingPlane"].update(type="array")
        metadata_schema["properties"]["Ophys"]["properties"]["TwoPhotonSeries"].update(type="array")

        metadata_schema["properties"]["Ophys"]["properties"]["Fluorescence"]["properties"]["roi_response_series"][
            "items"
        ]["required"] = list()
        metadata_schema["properties"]["Ophys"]["properties"]["ImageSegmentation"]["additionalProperties"] = True
        metadata_schema["properties"]["Ophys"]["properties"]["Fluorescence"]["properties"]["roi_response_series"].pop(
            "maxItems"
        )

        fill_defaults(metadata_schema, self.get_metadata())
        return metadata_schema

    def get_metadata(self):
        metadata = super().get_metadata()
        metadata.update(get_nwb_segmentation_metadata(self.segmentation_extractor))
        _ = metadata.pop("NWBFile")
        return metadata

    def run_conversion(self, nwbfile: NWBFile, metadata: dict, overwrite: bool = False):
        write_segmentation(self.segmentation_extractor, nwbfile=nwbfile, metadata=metadata, overwrite=overwrite)



"""Authors: Steffen Buergers."""
import os
import dateutil
import numpy as np
from pathlib import Path
from typing import Union

import spikeextractors as se
from pynwb import NWBFile
from pynwb.behavior import Position, SpatialSeries

from ..baserecordingextractorinterface import BaseRecordingExtractorInterface
from ..baselfpextractorinterface import BaseLFPExtractorInterface
from ....basedatainterface import BaseDataInterface
from ....tools.nwb_helpers import get_module
from ....utils import get_schema_from_method_signature, FilePathType


# Helper functions for AxonaRecordingExtractorInterface
def parse_generic_header(file_path: FilePathType, params: Union[list, set]):
    """
    Given a binary file with phrases and line breaks, enters the
    first word of a phrase as dictionary key and the following
    string (without linebreaks) as value. Returns the dictionary.

    INPUT
    file_path (str): .set file path and name.
    params (list or set): parameter names to search for.

    OUTPUT
    header (dict): dictionary with keys being the parameters that
                   were found & values being strings of the data.

    EXAMPLE
    parse_generic_header('myset_file.set', ['experimenter', 'trial_time'])
    """
    header = dict()
    if params is not None:
        params = set(params)
    with open(file_path, "rb") as f:
        for bin_line in f:
            if b"data_start" in bin_line:
                break
            line = bin_line.decode("cp1252").replace("\r\n", "").replace("\r", "").strip()
            parts = line.split(" ")
            key = parts[0]
            if params is None or key in params:
                header[key] = " ".join(parts[1:])
    return header


def read_axona_iso_datetime(set_file: FilePathType):
    """
    Creates datetime object (y, m, d, h, m, s) from .set file header
    and converts it to ISO 8601 format
    """
    with open(set_file, "r", encoding="cp1252") as f:
        for line in f:
            if line.startswith("trial_date"):
                date_string = line[len("trial_date") + 1 :].replace("\n", "")
            if line.startswith("trial_time"):
                time_string = line[len("trial_time") + 1 :].replace("\n", "")
    return dateutil.parser.parse(date_string + " " + time_string).isoformat()


class AxonaRecordingExtractorInterface(BaseRecordingExtractorInterface):
    """Primary data interface class for converting a AxonaRecordingExtractor"""

    RX = se.AxonaRecordingExtractor

    def __init__(self, file_path: FilePathType, verbose: bool = True):
        super().__init__(filename=file_path, verbose=verbose)
        self.source_data = dict(file_path=file_path, verbose=verbose)

    def get_metadata(self):

        # Extract information for specific parameters from .set file
        params_of_interest = ["experimenter", "comments", "duration", "sw_version"]
        set_file = self.source_data["file_path"].split(".")[0] + ".set"
        par = parse_generic_header(set_file, params_of_interest)

        # Extract information from AxonaRecordingExtractor
        elec_group_names = self.recording_extractor.get_channel_groups()
        unique_elec_group_names = set(elec_group_names)

        # Add available metadata
        metadata = super().get_metadata()
        metadata["NWBFile"] = dict(
            session_start_time=read_axona_iso_datetime(set_file),
            session_description=par["comments"],
            experimenter=[par["experimenter"]],
        )

        metadata["Ecephys"] = dict(
            Device=[
                dict(
                    name="Axona",
                    description="Axona DacqUSB, sw_version={}".format(par["sw_version"]),
                    manufacturer="Axona",
                ),
            ],
            ElectrodeGroup=[
                dict(
                    name=f"{group_name}",
                    location="",
                    device="Axona",
                    description=f"Group {group_name} electrodes.",
                )
                for group_name in unique_elec_group_names
            ],
        )

        return metadata


class AxonaUnitRecordingExtractorInterface(AxonaRecordingExtractorInterface):
    """Primary data interface class for converting a AxonaRecordingExtractor"""

    RX = se.AxonaUnitRecordingExtractor

    @classmethod
    def get_source_schema(cls):
        return dict(
            required=["file_path"],
            properties=dict(
                file_path=dict(
                    type="string",
                    format="file",
                    description="Path to Axona file.",
                ),
                noise_std=dict(type="number"),
            ),
            type="object",
        )

    def __init__(self, file_path: FilePathType, noise_std: float = 3.5):
        super().__init__(filename=file_path, noise_std=noise_std)
        self.source_data = dict(file_path=file_path, noise_std=noise_std)


# Helper functions for AxonaPositionDataInterface
def get_header_bstring(file: FilePathType):
    """
    Scan file for the occurrence of 'data_start' and return the header
    as byte string

    Parameters
    ----------
    file (str or path): file to be loaded

    Returns
    -------
    str: header byte content
    """
    header = b""
    with open(file, "rb") as f:
        for bin_line in f:
            if b"data_start" in bin_line:
                header += b"data_start"
                break
            else:
                header += bin_line
    return header


def read_bin_file_position_data(bin_file_path: FilePathType):
    """
    Read position data from Axona `.bin` file (if present).

    Parameters:
    -------
    bin_file_path (Path or Str):
        Full file_path of Axona file with any extension.

    Returns:
    -------
    np.array
        Columns are time (ms), X, Y, x, y, PX, px, tot_px, unused

    Notes:
    ------
    To obtain the correct column order we pairwise flip the 8 int16 columns
    described in the file format manual. In addition, note that `.bin` data is
    little endian (read right to left), as opposed to `.pos` file data, which is
    big endian.
    """
    pos_dt_se = np.dtype(
        [
            ("t", "<i4"),
            ("X", "<i2"),
            ("Y", "<i2"),
            ("x", "<i2"),
            ("y", "<i2"),
            ("PX", "<i2"),
            ("px", "<i2"),
            ("tot_px", "<i2"),
            ("unused", "<i2"),
        ]
    )

    bin_dt = np.dtype(
        [
            ("id", "S4"),
            ("packet", "<i4"),
            ("di", "<i2"),
            ("si", "<i2"),
            ("pos", pos_dt_se),
            ("ephys", np.byte, 384),
            ("trailer", np.byte, 16),
        ]
    )

    np_bin = np.memmap(
        filename=bin_file_path,
        dtype=bin_dt,
        mode="r",
    )

    # Only packets with the ADU2 flag contain position data
    pos_mask = np.where([np_bin["id"] == b"ADU2"])[1]
    pos_data = np_bin["pos"][pos_mask]

    # Rearrange columns of coordinates and pixels to conform with pos data
    # description in file format manual
    pos_data = np.vstack(
        (
            pos_data["Y"],
            pos_data["X"],
            pos_data["y"],
            pos_data["x"],
            pos_data["px"],
            pos_data["PX"],
            pos_data["unused"],
            pos_data["tot_px"],
        )
    ).T

    # Add timestamp as first column
    pos_data = np.hstack((pos_mask.reshape((-1, 1)), pos_data))

    # Create timestamps from position of samples in `.bin` file to ensure
    # alignment with ecephys data
    set_file = bin_file_path.split(".")[0] + ".set"
    sr_ecephys = int(parse_generic_header(set_file, ["rawRate"])["rawRate"])

    packets_per_ms = sr_ecephys / 3000
    pos_data[:, 0] = pos_data[:, 0] / packets_per_ms

    # Select only every second sample
    # Note that we do not lowpass filter, since other processing steps done by
    # TINT would no longer work properly.
    pos_data = pos_data[::2]

    return pos_data


def read_pos_file_position_data(pos_file_path: FilePathType):
    """
    Read position data from Axona `.pos` file.

    Parameters:
    -------
    pos_file_path (Path or Str):
        Full file_path of Axona file with any extension.

    Returns:
    -------
    np.array
        Columns are time (ms), X, Y, x, y, PX, px, tot_px, unused
    """

    pos_file_path = pos_file_path.split(".")[0] + ".pos"

    bytes_packet = 20
    footer_size = len("\r\ndata_end\r\n")
    header_size = len(get_header_bstring(pos_file_path))
    num_bytes = os.path.getsize(pos_file_path) - header_size - footer_size
    num_packets = num_bytes // bytes_packet

    pos_dt = np.dtype(
        [
            ("t", ">i4"),
            ("X", ">i2"),
            ("Y", ">i2"),
            ("x", ">i2"),
            ("y", ">i2"),
            ("PX", ">i2"),
            ("px", ">i2"),
            ("tot_px", ">i2"),
            ("unused", ">i2"),
        ]
    )

    pos_data = np.memmap(
        filename=pos_file_path,
        dtype=pos_dt,
        mode="r",
        offset=len(get_header_bstring(pos_file_path)),
        shape=(num_packets,),
    )

    # Convert structured memory mapped array to np array
    pos_data = np.vstack(
        (
            pos_data["X"],
            pos_data["Y"],
            pos_data["x"],
            pos_data["Y"],
            pos_data["PX"],
            pos_data["px"],
            pos_data["tot_px"],
            pos_data["unused"],
        )
    ).T

    # Create time column in ms assuming regularly sampled data starting from 0
    set_file = pos_file_path.split(".")[0] + ".set"
    dur_ecephys = float(parse_generic_header(set_file, ["duration"])["duration"])

    pos_data = np.hstack(
        (np.linspace(start=0, stop=dur_ecephys * 1000, num=pos_data.shape[0]).astype(int).reshape((-1, 1)), pos_data)
    )

    return pos_data


def get_position_object(file_path: FilePathType):
    """
    Read position data from .bin or .pos file and convert to
    pynwb.behavior.SpatialSeries objects. If possible it should always
    be preferred to read position data from the `.bin` file to ensure
    samples are locked to ecephys time courses.

    Parameters:
    ----------
    file_path (Path or Str):
        Full file_path of Axona file with any extension.

    Returns:
    -------
    position (pynwb.behavior.Position)
    """
    position = Position()

    position_channel_names = [
        "time(ms)",
        "X",
        "Y",
        "x",
        "y",
        "PX",
        "px",
        "px_total",
        "unused",
    ]

    if Path(file_path).suffix == ".bin":
        position_data = read_bin_file_position_data(file_path)
    else:
        position_data = read_pos_file_position_data(file_path)
    position_timestamps = position_data[:, 0]

    for ichan in range(0, position_data.shape[1]):

        spatial_series = SpatialSeries(
            name=position_channel_names[ichan],
            timestamps=position_timestamps,
            data=position_data[:, ichan],
            reference_frame="start of raw acquisition (.bin file)",
        )
        position.add_spatial_series(spatial_series)
    return position


class AxonaPositionDataInterface(BaseDataInterface):
    """Primary data interface class for converting Axona position data"""

    @classmethod
    def get_source_schema(cls):
        return get_schema_from_method_signature(cls.__init__)

    def __init__(self, file_path: str):
        super().__init__(filename=file_path)
        self.source_data(file_path=file_path)

    def run_conversion(self, nwbfile: NWBFile, metadata: dict):
        """
        Run conversion for this data interface.

        Parameters
        ----------
        nwbfile : NWBFile
        metadata : dict
        """
        file_path = self.source_data["file_path"]

        # Create or update processing module for behavioral data
        behavior_module = get_module(nwbfile=nwbfile, name="behavior", description="behavioral data")
        behavior_module.add(get_position_object(file_path))


# Helper functions for AxonaLFPDataInterface
def get_eeg_sampling_frequency(file_path: FilePathType):
    """
    Read sampling frequency from .eegX or .egfX file header.

    Parameters:
    -----------
    file_path : Path or str
        Full file_path of Axona `.eegX` or `.egfX` file.

    Returns:
    --------
    Fs : int
        Sampling frequency
    """
    Fs_entry = parse_generic_header(file_path, ["sample_rate"])
    Fs = int(float(Fs_entry.get("sample_rate").split(" ")[0]))

    return Fs


def read_eeg_file_lfp_data(file_path: FilePathType):
    """
    Read LFP data from Axona `.eegX` or `.egfX` file.

    Parameters:
    -------
    file_path (Path or Str):
        Full file_path of Axona `.eegX` or `.egfX` file.

    Returns:
    -------
    np.memmap (nobs x 1)
    """

    lfp_dtype = ">i1"
    footer_size = len("\r\ndata_end\r\n")
    header_size = len(get_header_bstring(file_path))
    num_bytes = os.path.getsize(file_path) - header_size - footer_size

    # .eeg files are int8, .egf files are int16
    if str(file_path).split(".")[1][0:3] == "egf":
        lfp_dtype = ">i2"
        num_bytes = num_bytes // 2
    eeg_data = np.memmap(
        filename=file_path,
        dtype=lfp_dtype,
        mode="r",
        offset=len(get_header_bstring(file_path)),
        shape=(1, num_bytes),
    )

    return eeg_data


def get_all_file_paths(file_path: FilePathType):
    """
    Read LFP file_paths of `.eeg` or `.egf` files in file_path's directory.
    E.g. if file_path='/my/directory/my_file.eeg', all .eeg channels will be
    appended to the output.

    Parameters:
    -----------
    file_path : path-like
        Full file_path of either .egg or .egf file

    Returns:
    --------
    path_list : list
        List of file_paths
    """

    suffix = Path(file_path).suffix[0:4]
    current_path = Path(file_path).parent

    path_list = [cur_path.name for cur_path in Path(file_path).parent.rglob("*" + suffix + "*")]

    return path_list


def read_all_eeg_file_lfp_data(file_path: FilePathType):
    """
    Read LFP data from all Axona `.eeg` or `.egf` files in file_path's directory.
    E.g. if file_path='/my/directory/my_file.eeg', all .eeg channels will be conactenated
    to a single np.array (chans x nobs). For .egf files substitude the file suffix.

    Parameters:
    -------
    file_path (Path or Str):
        Full file_path of Axona `.eeg` or `.egf` file.

    Returns:
    -------
    np.array (chans x obs)
    """

    file_path_list = get_all_file_paths(file_path)
    parent_path = Path(file_path).parent

    eeg_memmaps = list()
    sampling_rates = set()
    for fname in file_path_list:

        sampling_rates.add(get_eeg_sampling_frequency(parent_path / fname))

        eeg_memmaps.append(read_eeg_file_lfp_data(parent_path / fname))
    assert len(sampling_rates) < 2, "File headers specify different sampling rates. Cannot combine EEG data."

    eeg_data = np.concatenate(eeg_memmaps, axis=0)

    return eeg_data


class AxonaLFPDataInterface(BaseLFPExtractorInterface):
    """..."""

    RX = se.NumpyRecordingExtractor

    @classmethod
    def get_source_schema(cls):
        return dict(
            required=["file_path"],
            properties=dict(file_path=dict(type="string")),
            type="object",
            additionalProperties=False,
        )

    def __init__(self, file_path: FilePathType):
        super().__init__(
            timeseries=read_all_eeg_file_lfp_data(file_path), sampling_frequency=get_eeg_sampling_frequency(file_path)
        )
        self.source_data = dict(file_path=file_path)



from roiextractors import CnmfeSegmentationExtractor

from ..basesegmentationextractorinterface import BaseSegmentationExtractorInterface
from ....utils import FilePathType


class CnmfeSegmentationInterface(BaseSegmentationExtractorInterface):
    """Data interface for CnmfeRecordingInterface."""

    SegX = CnmfeSegmentationExtractor

    def __init__(self, file_path: FilePathType):
        super().__init__(file_path=file_path)



"""Authors: Cody Baker and Ben Dichter."""
import json
from jsonschema import validate
from typing import Optional, Dict
from pathlib import Path

from pynwb import NWBFile
from pynwb.file import Subject

from .tools.nwb_helpers import get_default_nwbfile_metadata, make_nwbfile_from_metadata, make_or_load_nwbfile
from .utils import (
    get_schema_from_hdmf_class,
    get_schema_for_NWBFile,
    dict_deep_update,
    get_base_schema,
    unroot_schema,
    fill_defaults,
)

from .utils.json_schema import NWBMetaDataEncoder


class NWBConverter:
    """Primary class for all NWB conversion classes."""

    data_interface_classes = None

    @classmethod
    def get_source_schema(cls):
        """Compile input schemas from each of the data interface classes."""
        source_schema = get_base_schema(
            root=True,
            id_="source.schema.json",
            title="Source data schema",
            description="Schema for the source data, files and directories",
            version="0.1.0",
        )
        for interface_name, data_interface in cls.data_interface_classes.items():
            source_schema["properties"].update({interface_name: unroot_schema(data_interface.get_source_schema())})
        return source_schema

    @classmethod
    def get_conversion_options_schema(cls):
        """Compile conversion option schemas from each of the data interface classes."""
        conversion_options_schema = get_base_schema(
            root=True,
            id_="conversion_options.schema.json",
            title="Conversion options schema",
            description="Schema for the conversion options",
            version="0.1.0",
        )
        for interface_name, data_interface in cls.data_interface_classes.items():
            conversion_options_schema["properties"].update(
                {interface_name: unroot_schema(data_interface.get_conversion_options_schema())}
            )
        return conversion_options_schema

    @classmethod
    def validate_source(cls, source_data: Dict[str, dict], verbose: bool = True):
        """Validate source_data against Converter source_schema."""
        cls._validate_source_data(source_data=source_data, verbose=verbose)

    def __init__(self, source_data: Dict[str, dict], verbose: bool = True):
        """Validate source_data against source_schema and initialize all data interfaces."""
        self.verbose = verbose
        self._validate_source_data(source_data=source_data, verbose=self.verbose)
        self.data_interface_objects = {
            name: data_interface(**source_data[name])
            for name, data_interface in self.data_interface_classes.items()
            if name in source_data
        }

    def get_metadata_schema(self):
        """Compile metadata schemas from each of the data interface objects."""
        metadata_schema = get_base_schema(
            id_="metadata.schema.json",
            root=True,
            title="Metadata",
            description="Schema for the metadata",
            version="0.1.0",
            required=["NWBFile"],
            properties=dict(NWBFile=get_schema_for_NWBFile(), Subject=get_schema_from_hdmf_class(Subject)),
        )
        for data_interface in self.data_interface_objects.values():
            interface_schema = unroot_schema(data_interface.get_metadata_schema())
            metadata_schema = dict_deep_update(metadata_schema, interface_schema)

        default_values = self.get_metadata()
        fill_defaults(metadata_schema, default_values)
        return metadata_schema

    def get_metadata(self):
        """Auto-fill as much of the metadata as possible. Must comply with metadata schema."""
        metadata = get_default_nwbfile_metadata()
        for interface in self.data_interface_objects.values():
            interface_metadata = interface.get_metadata()
            metadata = dict_deep_update(metadata, interface_metadata)
        return metadata

    def get_conversion_options(self):
        """Auto-fill as much of the conversion options as possible. Must comply with conversion_options_schema."""
        conversion_options = dict()
        for interface_name, interface in self.data_interface_objects.items():
            conversion_options[interface_name] = interface.get_conversion_options()
        return conversion_options

    def validate_metadata(self, metadata: Dict[str, dict]):
        """Validate metadata against Converter metadata_schema."""
        encoder = NWBMetaDataEncoder()
        # The encoder produces a serialiazed object so we de serialized it for comparison
        serialized_metadata = encoder.encode(metadata)
        decoded_metadata = json.loads(serialized_metadata)
        validate(instance=decoded_metadata, schema=self.get_metadata_schema())
        if self.verbose:
            print("Metadata is valid!")

    def validate_conversion_options(self, conversion_options: Dict[str, dict]):
        """Validate conversion_options against Converter conversion_options_schema."""
        validate(instance=conversion_options, schema=self.get_conversion_options_schema())
        if self.verbose:
            print("conversion_options is valid!")

    def _validate_source_data(self, source_data: Dict[str, dict], verbose: bool = True):
        validate(instance=source_data, schema=self.get_source_schema())
        if verbose:
            print("Source data is valid!")

    def run_conversion(
        self,
        nwbfile_path: Optional[str] = None,
        nwbfile: Optional[NWBFile] = None,
        metadata: Optional[dict] = None,
        overwrite: bool = False,
        conversion_options: Optional[dict] = None,
    ) -> NWBFile:
        """
        Run the NWB conversion over all the instantiated data interfaces.

        Parameters
        ----------
        nwbfile_path: FilePathType
            Path for where to write or load (if overwrite=False) the NWBFile.
            If specified, the context will always write to this location.
        nwbfile: NWBFile, optional
            An in-memory NWBFile object to write to the location.
        metadata: dict, optional
            Metadata dictionary with information used to create the NWBFile when one does not exist or overwrite=True.
        overwrite: bool, optional
            Whether or not to overwrite the NWBFile if one exists at the nwbfile_path.
            The default is False (append mode).
        verbose: bool, optional
            If 'nwbfile_path' is specified, informs user after a successful write operation.
            The default is True.
        conversion_options: dict, optional
            Similar to source_data, a dictionary containing keywords for each interface for which non-default
            conversion specification is requested.

        Returns
        -------
        nwbfile: NWBFile
            The in-memory NWBFile object after all conversion operations are complete.
        """
        if metadata is None:
            metadata = self.get_metadata()
        self.validate_metadata(metadata=metadata)

        if conversion_options is None:
            conversion_options = dict()
        default_conversion_options = self.get_conversion_options()
        conversion_options_to_run = dict_deep_update(default_conversion_options, conversion_options)
        self.validate_conversion_options(conversion_options=conversion_options_to_run)

        with make_or_load_nwbfile(
            nwbfile_path=nwbfile_path,
            nwbfile=nwbfile,
            metadata=metadata,
            overwrite=overwrite,
            verbose=self.verbose,
        ) as nwbfile_out:
            for interface_name, data_interface in self.data_interface_objects.items():
                data_interface.run_conversion(
                    nwbfile=nwbfile_out, metadata=metadata, **conversion_options_to_run.get(interface_name, dict())
                )
        return nwbfile_out



"""Authors: Cody Baker."""
import re
from typing import List
from warnings import warn


def decompose_f_string(f_string: str) -> (List[str], List[str]):
    """
    Decompose an f-string into the list of variable names and the separators between them.

    An f-string is any string that contains enclosed curly brackets around text.
    A variable is defined as the text expression within the enclosed curly brackets.
    The separators are the strings remnants that surround the variables.

    An example f-string and components would be: 'This is {an} f-string!', with variable 'an' and separators
    'This is ' and ' f-string!'.
    An instance of this example would be: 'This is definetely a good f-string!' with variable value 'definetely a good'.

    Example
    -------
        variable_names, separators = decompose_f_string(f_string="a/{x}b{y}/c{z}")
        # variable_names = ["x", "y", "z"]
        # separators = ["a/", "b", "/c"", ""]
    """
    matches = re.findall("{.*?}", f_string)  # {.*?} optionally matches any characters enclosed by curly brackets
    variable_names = [match.lstrip("{").rstrip("}") for match in matches]
    assert not any(
        (variable_name == "" for variable_name in variable_names)
    ), "Empty variable name detected in f-string! Please ensure there is text between all enclosing '{' and '}'."

    pattern = "^.*?{|}.*?{|}.*?$"
    # Description: patttern matches the all expressions outside of curly bracket enclosures
    #     .*?{   optionally matches any characters optionally before curly bracket opening
    #     |      logical 'or'
    #     }.*?{  between a curly bracket closure and opening
    #     |
    #     }.*?   after a closure
    separators = [x.rstrip("{").lstrip("}") for x in re.findall(pattern=pattern, string=f_string)]
    if any((separator == "" for separator in separators[1:-1])):
        warn(
            "There is an empty separator between two variables in the f-string! "
            "The f-string will not be uniquely invertible."
        )
    return variable_names, separators


def parse_f_string(string: str, f_string: str):
    """
    Given an instance of an f-string rule, extract the values of the variables specified by the f-string.

    Recovery of values is only possible in cases where the string instance is uniquely invertible,
    which requires at a minumum requires...
        1) Separators between all curly bracket enclosures, *e.g.*, '{var1}{var2}' is not allowed.
           An easy way to resolve this is to add a unique separator between them, *i.e.*, '{var1}-{var2}'.
        2) The separator character(s) cannot also occur within the variable values, *e.g.*, '{var1}b{var2}' on
           instance 'sub_01bsess_040122' where var1='sub_01 and' and var2='sess_040122'. Since the separator is a single
           character 'b' which also occurs in the instance of var1, it cannot be determined which occurence is the
           proper separator.

           Resolving this relies on choosing unique separators between variables in the f-string rule; either a single
           character that you know will never occur in any of your instances, or preferably a sequence of characters
           that would not occur together. In the example above, a simple separator of '-' would suffice, but if other
           instances might include that, such as var1='sub-05', then a sequential separator of '--' would work instead.

    Parameters
    ----------
    string : str
        An instance of the f-string rule.
    fstring : str
        String containing non-empty substrings enclosed by "{" and "}".
        These correspond to the names of variables thought to encode the actual filename string.
    """
    variable_names, separators = decompose_f_string(f_string=f_string)
    pattern = "^" + "(.+)".join(separators) + "$"  # (.+) matches any non-empty sequence of characters
    pattern_match = re.findall(pattern=pattern, string=string)
    assert pattern_match, "Unable to match f-string pattern to string! Please double check both structures."
    variable_values = pattern_match[0]
    for idx in range(len(variable_values) - 1):
        assert (
            separators[idx + 1] not in variable_values[idx]
        ), "Adjacent variable values contain the separator character! The f-string is not uniquely invertible."
    values = dict()
    for variable_name, variable_value in zip(variable_names, variable_values):
        if variable_value != values.get(variable_name, variable_value):
            raise ValueError(
                f"Duplicated variable placements for '{variable_name}' in f-string do not match in instance! "
                f"Expected '{values[variable_name]}' but found '{variable_value}'."
            )
        values.update({variable_name: variable_value})
    return values



"""Authors: Saksham Sharda, Cody Baker and Ben Dichter."""
from pathlib import Path
from typing import Optional
from warnings import warn

import psutil
import numpy as np
from hdmf.backends.hdf5.h5_utils import H5DataIO
from hdmf.data_utils import DataChunkIterator
from pynwb import NWBFile
from pynwb.image import ImageSeries
from tqdm import tqdm

from .movie_utils import VideoCaptureContext
from ....basedatainterface import BaseDataInterface
from ....tools.nwb_helpers import get_module
from ....utils import get_schema_from_hdmf_class, get_base_schema, calculate_regular_series_rate


def _check_duplicates(movies_metadata, file_paths):
    """
    Accumulates metadata for when multiple video files go in one ImageSeries container.

    Parameters
    ----------
    movies_metadata: List[Dict]
        The metadata corresponding to the movies should be organized as follow
                movies_metadata =[
                            dict(name="Video1", description="This is the first video.."),
                            dict(name="SecondVideo", description="Video #2 details..."),
                ]
    -------
    movies_metadata_unique: List[Dict]
        if metadata has common names (case when the user intends to put multiple video files
        under the same ImageSeries container), this removes the duplicate names.
    file_paths_list: List[List[str]]
        len(file_paths_list)==len(movies_metadata_unique)
    """
    keys_set = []
    movies_metadata_unique = []
    file_paths_list = []
    for n, movie in enumerate(movies_metadata):
        if movie["name"] not in keys_set:
            keys_set.append(movie["name"])
            file_paths_list.append([file_paths[n]])
            movies_metadata_unique.append(dict(movie))
        else:
            idx = keys_set.index(movie["name"])
            file_paths_list[idx].append(file_paths[n])

    return movies_metadata_unique, file_paths_list


class MovieInterface(BaseDataInterface):
    """Data interface for writing movies as ImageSeries."""

    def __init__(self, file_paths: list):
        """
        Create the interface for writing movies as ImageSeries.

        Parameters
        ----------
        file_paths : list of FilePathTypes
            Many movie storage formats segment a sequence of movies over the course of the experiment.
            Pass the file paths for this movies as a list in sorted, consecutive order.
        """
        super().__init__(file_paths=file_paths)

    def get_metadata_schema(self):
        metadata_schema = super().get_metadata_schema()
        image_series_metadata_schema = get_schema_from_hdmf_class(ImageSeries)
        # TODO: in future PR, add 'exclude' option to get_schema_from_hdmf_class to bypass this popping
        exclude = ["format", "conversion", "starting_time", "rate"]
        for key in exclude:
            image_series_metadata_schema["properties"].pop(key)
        metadata_schema["properties"]["Behavior"] = get_base_schema(tag="Behavior")
        metadata_schema["properties"]["Behavior"].update(
            required=["Movies"],
            properties=dict(
                Movies=dict(
                    type="array",
                    minItems=1,
                    items=image_series_metadata_schema,
                )
            ),
        )
        return metadata_schema

    def get_metadata(self):
        metadata = dict(
            Behavior=dict(
                Movies=[
                    dict(name=f"Video: {Path(file_path).stem}", description="Video recorded by camera.", unit="Frames")
                    for file_path in self.source_data["file_paths"]
                ]
            )
        )
        return metadata

    def run_conversion(
        self,
        nwbfile: NWBFile,
        metadata: dict,
        stub_test: bool = False,
        external_mode: bool = True,
        starting_times: Optional[list] = None,
        timestamps: Optional[list] = None,
        chunk_data: bool = True,
        module_name: Optional[str] = None,
        module_description: Optional[str] = None,
        compression: Optional[str] = "gzip",
        compression_options: Optional[int] = None,
    ):
        """
        Convert the movie data files to ImageSeries and write them in the NWBFile.
        Data is written in the ImageSeries container as RGB. [times, x, y, 3-RGB]

        Parameters
        ----------
        nwbfile : NWBFile
        metadata : dict
            Dictionary of metadata information such as names and description of each video.
            Metadata should be passed for each video file passed in the file_paths argument during __init__.
            If storing as 'external mode', then provide duplicate metadata for video files that go in the
            same ImageSeries container. len(metadata["Behavior"]["Movies"]==len(file_paths).
            Should be organized as follows:
                metadata = dict(
                    Behavior=dict(
                        Movies=[
                            dict(name="Video1", description="This is the first video.."),
                            dict(name="SecondVideo", description="Video #2 details..."),
                            ...
                        ]
                    )
                )
            and may contain most keywords normally accepted by an ImageSeries
            (https://pynwb.readthedocs.io/en/stable/pynwb.image.html#pynwb.image.ImageSeries).
             The list for the 'Movies' key should correspond one to one to the movie files in the file_paths list.
             If multiple movies need to be in the same ImageSeries, then supply the same value for "name" key.
             Storing multiple movies in the same ImageSeries is only supported if 'external_mode'=True.
        stub_test : bool, optional
            If True, truncates the write operation for fast testing. The default is False.
        external_mode : bool, optional
            ImageSeries in NWBFiles may contain either explicit movie data or file paths to external movie files. If
            True, this utilizes the more efficient method of merely encoding the file path linkage (recommended). For
            data sharing, the video files must be contained in the same folder as the NWBFile. If the intention of this
            NWBFile involves an upload to DANDI, the non-NWBFile types are not allowed so this flag would have to be
            set to False. The default is True.
        starting_times : list, optional
            List of start times for each movie. If unspecified, assumes that the movies in the file_paths list are in
            sequential order and are contiguous.
        timestamps : list, optional
            List of timestamps for the movies. If unspecified, timestamps are extracted from each movie data.
        chunk_data : bool, optional
            If True, uses a DataChunkIterator to read and write the movie, reducing overhead RAM usage at the cost of
            reduced conversion speed (compared to loading video entirely into RAM as an array). This will also force to
            True, even if manually set to False, whenever the video file size exceeds available system RAM by a factor
            of 70 (from compression experiments). Based on experiements for a ~30 FPS system of ~400 x ~600 color
            frames, the equivalent uncompressed RAM usage is around 2GB per minute of video. The default is True.
        module_name: str, optional
            Name of the processing module to add the ImageSeries object to. Default behavior is to add as acquisition.
        module_description: str, optional
            If the processing module specified by module_name does not exist, it will be created with this description.
            The default description is the same as used by the conversion_tools.get_module function.
        compression: str, optional
            Compression strategy to use for HFDataIO. For full list of currently supported filters, see
            https://docs.h5py.org/en/latest/high/dataset.html#lossless-compression-filters
        compression_options: int, optional
            Parameter(s) for compression filter. Currently only supports the compression level (integer from 0 to 9) of
            compression="gzip".
        """
        file_paths = self.source_data["file_paths"]

        if starting_times is not None:
            assert isinstance(starting_times, list) and all(
                [isinstance(x, float) for x in starting_times]
            ), "Argument 'starting_times' must be a list of floats."

        movies_metadata = metadata.get("Behavior", dict()).get("Movies", None)
        if movies_metadata is None:
            movies_metadata = self.get_metadata()["Behavior"]["Movies"]

        number_of_file_paths = len(file_paths)
        assert len(movies_metadata) == number_of_file_paths, (
            "Incomplete metadata "
            f"(number of metadata in movie {len(movies_metadata)})"
            f"is not equal to the number of file_paths {number_of_file_paths}"
        )

        movies_name_list = [movie["name"] for movie in movies_metadata]
        some_movie_names_are_not_unique = len(set(movies_name_list)) < len(movies_name_list)
        if some_movie_names_are_not_unique:
            assert external_mode, "For multiple video files under the same ImageSeries name, use exernal_mode=True."

        movies_metadata_unique, file_paths_list = _check_duplicates(movies_metadata, file_paths)

        if starting_times is not None:
            assert len(starting_times) == len(movies_metadata_unique), (
                f"starting times list length {len(starting_times)} must be equal to number of unique "
                f"ImageSeries {len(movies_metadata_unique)} \n"
                f"Movies metadata provided as input {movies_metadata} \n"
                f"starting times = {starting_times} \n"
                f"Image series after _check_duplicates {movies_metadata_unique}"
            )
        else:
            if len(movies_metadata_unique) == 1:
                warn("starting_times not provided, setting to 0.0")
                starting_times = [0.0]
            else:
                raise ValueError("provide starting times as a list of len " f"{len(movies_metadata_unique)}")

        for j, (image_series_kwargs, file_list) in enumerate(zip(movies_metadata_unique, file_paths_list)):

            if external_mode:
                with VideoCaptureContext(str(file_list[0])) as vc:
                    fps = vc.get_movie_fps()
                    if timestamps is None:
                        timestamps = starting_times[j] + vc.get_movie_timestamps()
                image_series_kwargs.update(
                    format="external",
                    external_file=file_list,
                )
            else:
                file = file_list[0]
                uncompressed_estimate = Path(file).stat().st_size * 70
                available_memory = psutil.virtual_memory().available
                if not chunk_data and not stub_test and uncompressed_estimate >= available_memory:
                    warn(
                        f"Not enough memory (estimated {round(uncompressed_estimate/1e9, 2)} GB) to load movie file as "
                        f"array ({round(available_memory/1e9, 2)} GB available)! Forcing chunk_data to True."
                    )
                    chunk_data = True
                with VideoCaptureContext(str(file)) as video_capture_ob:
                    if stub_test:
                        video_capture_ob.frame_count = 10
                    total_frames = video_capture_ob.get_movie_frame_count()
                    frame_shape = video_capture_ob.get_frame_shape()
                    timestamps = starting_times[j] + video_capture_ob.get_movie_timestamps()
                    fps = video_capture_ob.get_movie_fps()
                maxshape = (total_frames, *frame_shape)
                best_gzip_chunk = (1, frame_shape[0], frame_shape[1], 3)
                tqdm_pos, tqdm_mininterval = (0, 10)
                if chunk_data:
                    video_capture_ob = VideoCaptureContext(str(file))
                    if stub_test:
                        video_capture_ob.frame_count = 10
                    iterable = DataChunkIterator(
                        data=tqdm(
                            iterable=video_capture_ob,
                            desc=f"Copying movie data for {Path(file).name}",
                            position=tqdm_pos,
                            total=total_frames,
                            mininterval=tqdm_mininterval,
                        ),
                        iter_axis=0,  # nwb standard is time as zero axis
                        maxshape=maxshape,
                    )
                    data = H5DataIO(
                        iterable,
                        compression=compression,
                        compression_opts=compression_options,
                        chunks=best_gzip_chunk,
                    )
                else:
                    iterable = np.zeros(shape=maxshape, dtype="uint8")
                    with VideoCaptureContext(str(file)) as video_capture_ob:
                        if stub_test:
                            video_capture_ob.frame_count = 10
                        with tqdm(
                            desc=f"Reading movie data for {Path(file).name}",
                            position=tqdm_pos,
                            total=total_frames,
                            mininterval=tqdm_mininterval,
                        ) as pbar:
                            for n, frame in enumerate(video_capture_ob):
                                iterable[n, :, :, :] = frame
                                pbar.update(1)
                    data = H5DataIO(
                        DataChunkIterator(
                            tqdm(
                                iterable=iterable,
                                desc=f"Writing movie data for {Path(file).name}",
                                position=tqdm_pos,
                                mininterval=tqdm_mininterval,
                            ),
                            iter_axis=0,  # nwb standard is time as zero axis
                            maxshape=maxshape,
                        ),
                        compression="gzip",
                        compression_opts=compression_options,
                        chunks=best_gzip_chunk,
                    )
                image_series_kwargs.update(data=data)
            rate = calculate_regular_series_rate(series=timestamps)
            if rate is not None:
                if fps != rate:
                    warn(
                        f"The fps={fps} from movie data is unequal to the difference in "
                        f"regular timestamps. Using fps={rate} from timestamps instead.",
                        UserWarning,
                    )
                image_series_kwargs.update(starting_time=starting_times[j], rate=rate)
            else:
                image_series_kwargs.update(timestamps=timestamps)

            if module_name is None:
                nwbfile.add_acquisition(ImageSeries(**image_series_kwargs))
            else:
                get_module(nwbfile=nwbfile, name=module_name, description=module_description).add(
                    ImageSeries(**image_series_kwargs)
                )



"""Collection of modifications of HDMF functions that are to be tested/used on this repo until propagation upstream."""
from typing import Tuple

import numpy as np
from hdmf.data_utils import GenericDataChunkIterator as HDMFGenericDataChunkIterator


class GenericDataChunkIterator(HDMFGenericDataChunkIterator):
    def _get_default_buffer_shape(self, buffer_gb: float = 1.0) -> Tuple[int]:
        num_axes = len(self.maxshape)
        chunk_bytes = np.prod(self.chunk_shape) * self.dtype.itemsize
        assert buffer_gb > 0, f"buffer_gb ({buffer_gb}) must be greater than zero!"
        assert (
            buffer_gb >= chunk_bytes / 1e9
        ), f"buffer_gb ({buffer_gb}) must be greater than the chunk size ({chunk_bytes / 1e9})!"
        assert all(
            np.array(self.chunk_shape) > 0
        ), f"Some dimensions of chunk_shape ({self.chunk_shape}) are less than zero!"

        maxshape = np.array(self.maxshape)

        # Early termination condition
        if np.prod(maxshape) * self.dtype.itemsize / 1e9 < buffer_gb:
            return tuple(self.maxshape)

        buffer_bytes = chunk_bytes
        axis_sizes_bytes = maxshape * self.dtype.itemsize
        smallest_chunk_axis, second_smallest_chunk_axis, *_ = np.argsort(self.chunk_shape)
        target_buffer_bytes = buffer_gb * 1e9

        # If the smallest full axis does not fit within the buffer size, form a square along the two smallest axes
        sub_square_buffer_shape = np.array(self.chunk_shape)
        if min(axis_sizes_bytes) > target_buffer_bytes:
            k1 = np.floor((target_buffer_bytes / chunk_bytes) ** 0.5)
            for axis in [smallest_chunk_axis, second_smallest_chunk_axis]:
                sub_square_buffer_shape[axis] = k1 * sub_square_buffer_shape[axis]
            return tuple(sub_square_buffer_shape)

        # Original one-shot estimation has good performance for certain shapes
        chunk_to_buffer_ratio = buffer_gb * 1e9 / chunk_bytes
        chunk_scaling_factor = np.floor(chunk_to_buffer_ratio ** (1 / num_axes))
        unpadded_buffer_shape = [
            np.clip(a=int(x), a_min=self.chunk_shape[j], a_max=self.maxshape[j])
            for j, x in enumerate(chunk_scaling_factor * np.array(self.chunk_shape))
        ]

        unpadded_buffer_bytes = np.prod(unpadded_buffer_shape) * self.dtype.itemsize

        # Method that starts by filling the smallest axis completely or calculates best partial fill
        padded_buffer_shape = np.array(self.chunk_shape)
        chunks_per_axis = np.ceil(maxshape / self.chunk_shape)
        small_axis_fill_size = chunk_bytes * min(chunks_per_axis)
        full_axes_used = np.zeros(shape=num_axes, dtype=bool)
        if small_axis_fill_size <= target_buffer_bytes:
            buffer_bytes = small_axis_fill_size
            padded_buffer_shape[smallest_chunk_axis] = self.maxshape[smallest_chunk_axis]
            full_axes_used[smallest_chunk_axis] = True
        for axis, chunks_on_axis in enumerate(chunks_per_axis):
            if full_axes_used[axis]:  # If the smallest axis, skip since already used
                continue
            if chunks_on_axis * buffer_bytes <= target_buffer_bytes:  # If multiple axes can be used together
                buffer_bytes *= chunks_on_axis
                padded_buffer_shape[axis] = self.maxshape[axis]
            else:  # Found an axis that is too large to use with the rest of the buffer; calculate how much can be used
                k3 = np.floor(target_buffer_bytes / buffer_bytes)
                padded_buffer_shape[axis] *= k3
                break
        padded_buffer_bytes = np.prod(padded_buffer_shape) * self.dtype.itemsize

        if padded_buffer_bytes >= unpadded_buffer_bytes:
            return tuple(padded_buffer_shape)
        else:
            return tuple(unpadded_buffer_shape)


class SliceableDataChunkIterator(GenericDataChunkIterator):
    """
    Generic data chunk iterator that works for any memory mapped array, such as a np.memmap or an h5py.Dataset
    """

    def __init__(self, data, **kwargs):
        self.data = data
        super().__init__(**kwargs)

    def _get_dtype(self) -> np.dtype:
        return self.data.dtype

    def _get_maxshape(self) -> tuple:
        return self.data.shape

    def _get_data(self, selection: Tuple[slice]) -> np.ndarray:
        return self.data[selection]
